/* ************************************************************************** 
   Hearder file for the matlab mex wrapper of the map function in the Planar Inference code of Nic Schraudolph
   -- Dhruv Batra (batradhruv -at- cmu.edu) 
   ************************************************************************** */

#ifndef MEX_BLOSSOM_HPP
#define MEX_BLOSSOM_HPP

// Hack to get access to private members of base classes from derived classes
#define private protected

#include "blossom.hpp"
#include <ctime>

#include "utils.hpp"
#include "graph.h"

#include <fstream>
#include <string>
#include <map>

//#define DBL2INT_MAX 999999998.0
#define DBL2INT_MAX (PM_INFTY/1024)
#define DBL2INT_INF (PM_INFTY/8)
//#define DBL2INT_INF (PM_INFTY/512)

// Macros for sub2ind
//#define SUB2IND(r_ind, c_ind) (c_ind*2 + r_ind)
inline size_t SUB2IND(size_t r_ind, size_t c_ind, bool swap) 
{
  if (!swap)
    return(c_ind*2 + r_ind);
  else 
    return(r_ind*2 + c_ind);
}

namespace Isinf
{
  using namespace std;
  
time_t start,end;
  ///////////////////////////////////////
  // mexGraph is a class derived from the Graph class in graph.hpp
  // This class contains some functions not present in Graph, useful
  // for the mex wrappers.
  ///////////////////////////////////////
  class mexGraph : public BlossomGraph {

  public:
    int verbose_level;

    mexGraph() : BlossomGraph(), verbose_level(0) {};
    mexGraph(istream& is) : BlossomGraph(is), verbose_level(0) {};
    mexGraph(double *xvalues,int n_nodes,int max_degree); // build graph from embedding matrix
    mexGraph(double *xvalues,int n_edges);                // build graph from edge_list matrix
    void blossom_minmarg(double *min_marg, std::vector<double>& ew, const uint32_t *di);
    void blossom_ne_mm(double *node_mm, double *edge_mm, double *edge_list, std::vector<double>& ew, const uint32_t *di);

  protected:
    typedef PerfectMatching::REAL REAL;
    double run_matching(REAL e_sum, double scale2);
    void matching2states(uint32_t* states, double* en, vector<double>& ew, vector<bool>& bicon);
    void errCheck(int k);

    typedef std::pair<size_t,size_t> Edge;
    
    // set up map from node pair to edge index
    typedef std::pair<Edge, size_t> EdgeMapEntry;
    typedef std::map<Edge, size_t> EdgeMap;
    EdgeMap edgemap;
  };
  

  ///////////////////////////////////////
  // build graph from edge_list matrix
  ///////////////////////////////////////
  inline mexGraph::mexGraph(double *xvalues,int n_edges) : BlossomGraph(), verbose_level(0)
  {
    int n_nodes;
    int row,col,i,j;
    char s[50];
    int v1,v2;


    std::vector<Edge> edges_embed;
    size_t max = 0;
    int n = 0;

    for (i=0; i<n_edges; i++)
      {
	v1 = (int)ARR_CM(xvalues,i,0,n_edges);
	v2 = (int)ARR_CM(xvalues,i,1,n_edges);
	
	if (v1 == v2) { // self-loop! 
	  edgemap.clear();
	  sprintf(s,"error at node %d: self-loops not allowed\n",v1);
	  mexErrMsgTxt(s);
	}
	
	if (v1 > v2) std::swap(v1, v2);
	if (v2 > max) max = (size_t) v2;
	
	Edge e(v1, v2);
	
	edges_embed.push_back(e);
	
	if (!edgemap.insert(EdgeMapEntry(e, n++)).second) {
	  edgemap.clear();
	  sprintf(s,"duplicate edge (%d) between nodes %d and %d\n", i, v1, v2);
	  mexErrMsgTxt(s);
	}
      }
    if (edges_embed.size() < 2){      
      edgemap.clear();
      mexErrMsgTxt("less than 2 valid edges found\n");
    }
    else if (edges_embed.size() > 3*(++max) - 6) {
      // edge limit for planar graphs
      errCheck(NONPLANAR);
    }
    
    // create graph in planarity code
    n_nodes = max;
    graphP g = gp_New();
    gp_InitGraph(g, max);
    for (std::vector<Edge>::iterator i = edges_embed.begin(); i < edges_embed.end(); ++i)
      errCheck(gp_AddEdge(g, i->first, 0, i->second, 0));
    
    //writeAdjList(g);
    // timed call to planarity code
    Isinf::CPUtimer *t = new Isinf::CPUtimer;
    n = gp_Embed(g, EMBEDFLAGS_PLANAR);
    delete t;

    // report result
    if (n != OK)
      {
	gp_Free(&g);
	errCheck(n);
      }
    
    //printf("n_edges = %d, max = %d\n",edges_embed.size(),max);
    gp_SortVertices(g);

    // Copy information off of g and into mexGraph
    size_t n2;
    Edgelet* e2;
    for (i = 0; i < g->N; ++i)
      {
	j = g->G[i].link[1];
	while (j >= g->N)
	  {
	    Edge e(std::min(i, g->G[j].v), std::max(i, g->G[j].v));

	    EdgeMap::iterator it(edgemap.find(e));
	    if (it == edgemap.end()) {
	      sprintf(s,"can't find edge (%d,%d)\n",i,g->G[j].v);
	      edgemap.clear();
	      mexErrMsgTxt(s);
	    }
	    
	    //std::cout << it->second << ' ';
	    n2 = (size_t) it->second;
	    //change e-> e2, j->n2
	    e2 = new Edgelet;
	    
	    e2->index = n2;
	    if (n2 >= edges.size()) edges.resize(n2+1);
	    if (edges[n2]) {
	      e2->other = edges[n2];
	      assert(!edges[n2]->other);
	      edges[n2]->other = e2;
	      ++nedges;
	    }
	    else edges[n2] = e2;
	    
	    e2->node = i;
	    if (i >= nodes.size()) nodes.resize(i+1);
	    if (nodes[i].edges) {
	      e2->prev = nodes[i].edges->prev;
	      e2->prev->next = e2;
	      e2->next = nodes[i].edges;
	      nodes[i].edges->prev = e2;
	    }
	    else nodes[i].edges = e2->prev = e2->next = e2;


	    j = g->G[j].link[1];
	  }
      
      }
    genus();


    //free variables
    gp_Free(&g);
    edgemap.clear();

  }


  ///////////////////////////////////////
  // build graph from embedding matrix
  ///////////////////////////////////////
  inline mexGraph::mexGraph(double *xvalues,int n_nodes,int max_degree) : BlossomGraph(), verbose_level(0)
  {
    size_t i = 0, j, k;
    string s;
    Edgelet* e;
    int row,col,counter;
    
    // TODO: pre-assign space for nodes and edges.

    for (row=0; row<n_nodes; row++) 
      {
	for (col=0; col<max_degree; col++) 
	  {
	    j = (size_t) ARR_CM(xvalues,row,col,n_nodes);
	    if (j==-1) 
	      break;
	      
	    e = new Edgelet;
	      
	    e->index = j;
	    if (j >= edges.size()) edges.resize(j+1);
	    if (edges[j]) {
	      e->other = edges[j];
	      assert(!edges[j]->other);
	      edges[j]->other = e;
	      ++nedges;
	    }
	    else edges[j] = e;
	      
	    e->node = i;
	    if (i >= nodes.size()) nodes.resize(i+1);
	    if (nodes[i].edges) {
	      e->prev = nodes[i].edges->prev;
	      e->prev->next = e;
	      e->next = nodes[i].edges;
	      nodes[i].edges->prev = e;
	    }
	    else nodes[i].edges = e->prev = e->next = e;
	  }
	++i;
      }

    genus();
  }
  
  ///////////////////////////////////////
  // Compute Node Min-Marginals Only.
  ///////////////////////////////////////
  inline void mexGraph::blossom_minmarg(double *min_marg, std::vector<double>& ew, const uint32_t *di)
  {
    assert(!pm); // pm should not be defined at this point
    edgelen.clear();

    size_t k;
    double psum = 0.0, nsum = 0.0, pm_cost, cut_cost;
    REAL e_sum=0;
    int n;

    int n_nodes = nodes.size();
    int n_edges = ew.size();
    assert(n_edges > 0);

    for (k = 0; k != n_edges; ++k)
      {
	if (ew[k] > 0.0) 
	    psum += ew[k];
	else  nsum -= ew[k];
      }

    // BUGBUG: this depends on INTEGER_MAX def'n in match.c
    //const double max = DBL2INT_MAX/std::max(psum, nsum);
    scale = -DBL2INT_MAX/( (psum>nsum)?psum:nsum );
    double scale2 = -DBL2INT_MAX/( (psum>nsum)?psum:nsum );

    //printf("psum = %lf, nsum = %lf, max = %lf\n",psum,nsum,scale);

    // construct dual for blossom
    vector<bool> bicon(nedges, true);
    blossom_dual(bicon);
    edgelen.reserve(edgelist.size()/2);
    
    size_t s = edgelist.size()/2;
    assert( s > n_edges );
    for (k=0; k != s; ++k)
      if (k<n_edges) {
	if (IS_INT)
	  edgelen.push_back(REAL(scale2*ew[k]));
	else
	  edgelen.push_back(REAL(scale*ew[k]));
	e_sum -= edgelen[k];
      }
      else
	edgelen.push_back(0);
    

    // Set up Perfect Matching instance
    pm = new PerfectMatching(2*nedges, edgelist.size()/2);
    pm->options.verbose = false;
    pm->options.dual_greedy_update_option = 0;
    
    for (k = 0; k < (edgelist.size()/2); ++k) 
      n = pm->AddEdge(edgelist[2*k], edgelist[2*k + 1], edgelen[k]);
    
    if (verbose_level == 1)
      {printf("DP-Cut: First Cut...\t");fflush(stdout);}
    
    // Get unconstrained Perfect Matching
    pm->Solve();
    
    // call matching2states
    std::vector<uint32_t> states(n_nodes-1,0);
    matching2states(&states[0], &cut_cost, ew, bicon);
    
    //printf("cut_cost = %lf, e_sum = %lf\n", cut_cost, e_sum);
    
    // Store this min-cut cost in appropriate places
    for (k=0; k!=(n_nodes-1); ++k)
      ARR_CM(min_marg,states[k],k,2) = cut_cost;

    // Constrained cuts and edge updates
    pm->StartUpdate();
    for (k=0; k!=(n_nodes-1); ++k)
      {
	
	if (states[k]) // State 1 => s-k edge CUT => edge NOT in matching => force to be in matching by -INFTY cost
	  pm->UpdateCost(di[k],-DBL2INT_INF); 
	else 
	  pm->UpdateCost(di[k],DBL2INT_INF);
	
	pm->FinishUpdate();
	
	// Get new matching
	pm->Solve();
	
	// Compute min-cut cost
	pm_cost = ComputePerfectMatchingCost(2*nedges, edgelist.size()/2, &edgelist[0], &edgelen[0], pm);
	if (IS_INT)
	  cut_cost = (e_sum + pm_cost)/(-scale2);
	else
	  cut_cost = (e_sum + pm_cost)/(-scale);
	//printf("pm_cost: %lf\n",pm_cost);


	// Save this cost at the right place
	ARR_CM(min_marg,1-states[k],k,2) = cut_cost;
	
	if (k!=(n_nodes-2)) { // If not the last node remove the inf costs
	  pm->StartUpdate();
	  if (states[k]) 
	    pm->UpdateCost(di[k],+DBL2INT_INF);
	  else 
	    pm->UpdateCost(di[k],-DBL2INT_INF);
	}
	
	if (verbose_level == 1)
	  if ((k%1000) == 0)
	    {printf("Node %d...\t",k); fflush(stdout);}
      }
  }



  ///////////////////////////////////////
  // Compute Node and Edge Min-Marginals
  ///////////////////////////////////////
  inline void mexGraph::blossom_ne_mm(double *node_mm, double *edge_mm, double *edge_list, std::vector<double>& ew, const uint32_t *di)
  {
    assert(!pm); // pm should not be defined at this point
    edgelen.clear();

    size_t k;
    double psum = 0.0, nsum = 0.0, pm_cost, cut_cost, ecut_cost;
    REAL e_sum=0;
    int n;
    uint32_t v1, v2;

    int n_nodes = nodes.size();
    int n_edges = ew.size();
    int n_gedges = n_edges - n_nodes+1;
    assert(n_edges > 0);

    bool up_started = false;

    for (k = 0; k != n_edges; ++k)
      {
	if (ew[k] > 0.0) 
	    psum += ew[k];
	else  nsum -= ew[k];
      }

    // BUGBUG: this depends on INTEGER_MAX def'n in match.c
    //const double max = DBL2INT_MAX/std::max(psum, nsum);
    scale = -DBL2INT_MAX/( (psum>nsum)?psum:nsum );
    double scale2 = -DBL2INT_MAX/( (psum>nsum)?psum:nsum );

    //printf("psum = %lf, nsum = %lf, max = %lf\n",psum,nsum,scale);

    // construct dual for blossom
    vector<bool> bicon(nedges, true);
    blossom_dual(bicon);
    edgelen.reserve(edgelist.size()/2);
    
    size_t s = edgelist.size()/2;
    assert( s > n_edges );
    for (k=0; k != s; ++k)
      if (k<n_edges) {
	if (IS_INT)
	  edgelen.push_back(REAL(scale2*ew[k]));
	else
	  edgelen.push_back(REAL(scale*ew[k]));
	e_sum -= edgelen[k];
      }
      else
	edgelen.push_back(0);
    

    // Set up Perfect Matching instance
    pm = new PerfectMatching(2*nedges, edgelist.size()/2);
    pm->options.verbose = false;
    pm->options.dual_greedy_update_option = 0;
    
    for (k = 0; k < (edgelist.size()/2); ++k) 
      n = pm->AddEdge(edgelist[2*k], edgelist[2*k + 1], edgelen[k]);
    
    if (verbose_level == 1)
      {printf("DP-Cut: First Cut...\t");fflush(stdout);}
    
    // Get unconstrained Perfect Matching
    pm->Solve();
    
    // call matching2states
    std::vector<uint32_t> states(n_nodes-1,0);
    std::vector<uint32_t> states2(n_nodes-1,0);
    std::vector<uint32_t> states3(n_nodes-1,0);
    matching2states(&states[0], &cut_cost, ew, bicon);
    
    //printf("cut_cost = %lf, e_sum = %lf\n", cut_cost, e_sum);
    
    // Store this min-cut cost in appropriate places in node_mm
    for (k=0; k!=(n_nodes-1); ++k)
      ARR_CM(node_mm,states[k],k,2) = cut_cost;
    
    // Datastructure to keep track of edges
    uint8_t *edge_marker = new uint8_t[n_gedges];
    memset(edge_marker,0,n_gedges*sizeof(uint8_t)); // initialize to 0
    
    // Store this min-cut cost in appropriate places in edge_mm
    for (k=0; k!=(n_gedges); ++k) {
      v1 =  (uint32_t) ARR_CM(edge_list,k,0,n_edges); v2 = (uint32_t) ARR_CM(edge_list,k,1,n_edges); 
      ARR_CM(edge_mm,SUB2IND(states[v1],states[v2],0),k,4) = cut_cost;
      
      // mark this state
      edge_marker[k] = (edge_marker[k] | (1<<SUB2IND(states[v1],states[v2],0)) );
    }

    int multiplier = 2;
    // Constrained cuts and edge updates
    pm->StartUpdate();
    for (k=0; k!=(n_nodes-1); ++k)
      {
	if (states[k]) // State 1 => s-k edge CUT => edge NOT in matching => force to be in matching by -INFTY cost
	  pm->UpdateCost(di[k],-DBL2INT_INF); 
	else 
	  pm->UpdateCost(di[k],DBL2INT_INF);
	
	pm->FinishUpdate();
	
	// Get new matching
	pm->Solve(); 

	// Compute min-cut cost
	pm_cost = ComputePerfectMatchingCost(2*nedges, edgelist.size()/2, &edgelist[0], &edgelen[0], pm);
	if (IS_INT)
	  cut_cost = (e_sum + pm_cost)/(-scale2);
	else
	  cut_cost = (e_sum + pm_cost)/(-scale);
	//cut_cost = run_matching();
	
	//printf("pm_cost: %lf\n",pm_cost);
	matching2states(&states2[0], &cut_cost, ew, bicon);

	// Save this cost at the right place in node_mm
	ARR_CM(node_mm,1-states[k],k,2) = cut_cost;
	
	// Loop on edges containing this node
	size_t counter = 0;
	if (Edgelet* e = nodes[k].edges) do
	  {
	    size_t v2 = k; 
	    size_t v1 = e->other->node;
	    bool swap_nodes = false;

	    if (v1 == (n_nodes-1)) {
	      e = e->next;
	      continue;
	    }
	    counter++;

	    // define swap variable based on ordering in edge_list
	    if (v1 == (size_t) ARR_CM(edge_list,e->index,1,n_edges))
		swap_nodes = true;
	    else
	      {assert(v1 == (size_t) ARR_CM(edge_list,e->index,0,n_edges));}
	    
	    // Save this cost at the right place in edge_mm
	    ARR_CM(edge_mm,SUB2IND(states2[v1],states2[v2],swap_nodes),e->index,4) = cut_cost;
	    // mark this state
	    edge_marker[e->index] = (edge_marker[e->index] | (1<<SUB2IND(states2[v1],states2[v2],swap_nodes)) );
	    
	    if ((edge_marker[e->index] & 0x80)==0) // Not seen this edge before, continue looping on edges
	      {
		edge_marker[e->index] = (edge_marker[e->index] | 0x80); // mark edge
		e = e->next;
		continue;
	      }
	    
	    // count empty places
	    uint8_t v = edge_marker[e->index];
	    uint8_t c; // accum
	    for (c = 0; v; c++)
	      v &= v - 1; // clear the least significant bit set
	    c = 4-c+1;
	    
	    assert((c==1)||(c==2));
	    if (!up_started)
	      {
		pm->StartUpdate();
		up_started = true;
	      }

	    if (c==1) // 1 empty spot in edge_mm
	      {
		bool s_ev1 = ((edge_marker[e->index] & 0x0A) !=0x0A); //1010
		bool s_ev2 = ((edge_marker[e->index] & 0x0C) !=0x0C); //1100

		if (swap_nodes)
		  swap(s_ev1,s_ev2);
		
		assert(s_ev1 != states2[v1]); // empty spot can't have same state as that of v1
		
		// Put +- inf on v1 edge
		if (states2[v1]) // State 1 => s-k edge CUT => edge NOT in matching => force to be in matching by -INFTY cost
		  pm->UpdateCost(di[v1],-DBL2INT_INF); 
		else 
		  pm->UpdateCost(di[v1],DBL2INT_INF);
		
		// Check if in the empty spot state of v2 has also changed
		if (s_ev2 != states2[v2])
		  {
		    //printf("Never seen before case: State of node %d has also changed (%d to %d)\n",v2,states2[v2],(int)s_ev2);
		    // If so, change cost on (o,v2) edge too
		    if (states2[v2]) // State 1 => s-k edge CUT => edge NOT in matching => force to be in matching by -INFTY cost
		      pm->UpdateCost(di[v2],-multiplier*DBL2INT_INF);  // 1 to cancel, 1 to force
		    else 
		      pm->UpdateCost(di[v2],multiplier*DBL2INT_INF); 
		  }
		    
		pm->FinishUpdate(); 
		// Get new matching
		pm->Solve();

		// Compute min-cut cost
		pm_cost = ComputePerfectMatchingCost(2*nedges, edgelist.size()/2, &edgelist[0], &edgelen[0], pm);
		if (IS_INT)
		  ecut_cost = (e_sum + pm_cost)/(-scale2);
		else
		  ecut_cost = (e_sum + pm_cost)/(-scale);
		    
		//matching2states(&states3[0], &ecut_cost, ew, bicon);
		
		// Save this cost at the right place in edge_mm
		if (s_ev2 != states2[v2])
		  {
		    ARR_CM(edge_mm,SUB2IND(1-states2[v1],1-states2[v2],swap_nodes),e->index,4) = ecut_cost;
		    // mark state
		    edge_marker[e->index] = (edge_marker[e->index] | (1<<SUB2IND(1-states2[v1],1-states2[v2],swap_nodes)) );
		  }
		else
		  {
		    ARR_CM(edge_mm,SUB2IND(1-states2[v1],states2[v2],swap_nodes),e->index,4) = ecut_cost;
		    // mark state
		    edge_marker[e->index] = (edge_marker[e->index] | (1<<SUB2IND(1-states2[v1],states2[v2],swap_nodes)) );
		  }
		
		// remove these weights
		pm->StartUpdate();  
		if (states2[v1]) 
		  pm->UpdateCost(di[v1],+DBL2INT_INF);
		else 
		  pm->UpdateCost(di[v1],-DBL2INT_INF);
		// Check if in the empty spot state of v2 has also changed
		if (s_ev2 != states2[v2])
		  {
		    if (states2[v2]) 
		      pm->UpdateCost(di[v2],+multiplier*DBL2INT_INF);
		    else 
		      pm->UpdateCost(di[v2],-multiplier*DBL2INT_INF);
		  }
	      } // End of 1 empty spot stuff
	    else  // 2 empty spots in edge_mm
	      {
		bool ondiag = ((edge_marker[e->index] & 0x09) ==0); // 1001
		if (!ondiag)
		  assert((edge_marker[e->index] & 0x06) ==0); // 0110
		
		// Put +-inf on (v1,v2) edge
		if (ondiag) // Empty on diag => Need to couple nodes => (v1,v2) edges NOT in cut => (v1,v2) edge in matching => set -inf
		  pm->UpdateCost(e->index,-DBL2INT_INF); 
		else 
		  pm->UpdateCost(e->index,DBL2INT_INF);
		
		// 2-constrained cuts
		for (int i=0; i!=2; ++i)
		  {
		    if (i==1)
		      {
			pm->StartUpdate(); 
			// add +-inf weight on (o,v2) edge
			if (states2[v2]) // State 1 => s-k edge CUT => edge NOT in matching => force to be in matching by -INFTY cost
			  pm->UpdateCost(di[v2],-multiplier*DBL2INT_INF);  // 1 to cancel, 1 to force
			else 
			  pm->UpdateCost(di[v2],multiplier*DBL2INT_INF);
		      }
		    
		    pm->FinishUpdate();
		    // Get new matching
		    pm->Solve();

		    // Compute min-cut cost
		    pm_cost = ComputePerfectMatchingCost(2*nedges, edgelist.size()/2, &edgelist[0], &edgelen[0], pm);
		    if (IS_INT)
		      ecut_cost = (e_sum + pm_cost)/(-scale2);
		    else
		      ecut_cost = (e_sum + pm_cost)/(-scale);
		    
		    //matching2states(&states3[0], &ecut_cost, ew, bicon);

		    // Save this cost at the right place in edge_mm
		    if (i==0)
		      if (ondiag)
			{
			  ARR_CM(edge_mm,SUB2IND(states2[v2],states2[v2],swap_nodes),e->index,4) = ecut_cost;
			  // mark state
			  edge_marker[e->index] = (edge_marker[e->index] | (1<<SUB2IND(states2[v2],states2[v2],swap_nodes)) );
			}
		      else
			{
			  ARR_CM(edge_mm,SUB2IND(1-states2[v2],states2[v2],swap_nodes),e->index,4) = ecut_cost;
			  // mark state
			  edge_marker[e->index] = (edge_marker[e->index] | (1<<SUB2IND(1-states2[v2],states2[v2],swap_nodes)) );
			}
		    else
		      if (ondiag)
			{
			  ARR_CM(edge_mm,SUB2IND(1-states2[v2],1-states2[v2],swap_nodes),e->index,4) = ecut_cost;
			  // mark state
			  edge_marker[e->index] = (edge_marker[e->index] | (1<<SUB2IND(1-states2[v2],1-states2[v2],swap_nodes)) );
			}
		      else
			{
			  ARR_CM(edge_mm,SUB2IND(states2[v2],1-states2[v2],swap_nodes),e->index,4) = ecut_cost;
			  // mark state
			  edge_marker[e->index] = (edge_marker[e->index] | (1<<SUB2IND(states2[v2],1-states2[v2],swap_nodes)) );
			}
		  } // End of for loop for 2 constrainted cuts
		
		// remove these weights
		pm->StartUpdate(); 
		if (ondiag) 
		  pm->UpdateCost(e->index,+DBL2INT_INF);
		else 
		  pm->UpdateCost(e->index,-DBL2INT_INF);
		// remove +- inf weight on (o,v2) edge
		if (states2[v2]) 
		  pm->UpdateCost(di[v2],+multiplier*DBL2INT_INF);
		else 
		  pm->UpdateCost(di[v2],-multiplier*DBL2INT_INF);
		
	      } // End of 2 empty spots stuff

	    edge_marker[e->index] = (edge_marker[e->index] | 0x80); // mark edge
	    e = e->next;
	  }
	while (e != nodes[k].edges); // End of loop on edges of node k
	
	if (k!=(n_nodes-2)) { // If not the last node remove the inf costs
	  if (!up_started) 
	    pm->StartUpdate(); 
	  up_started = false;
	  if (states[k]) 
	    pm->UpdateCost(di[k],+DBL2INT_INF);
	  else 
	    pm->UpdateCost(di[k],-DBL2INT_INF);
	}

	if (up_started) 
	  pm->FinishUpdate();
	
	if (verbose_level == 1)
	  if ((k%1000) == 0)
	    {printf("Node %d...\t",k); fflush(stdout);}
      } // End of loop on nodes 
    

    // release edge_marker
    delete[] edge_marker;
  }


  
  inline double mexGraph::run_matching(REAL e_sum, double scale2)
  {
    // Get new matching
    pm->Solve();
    
    // Compute min-cut cost
    double pm_cost = ComputePerfectMatchingCost(2*nedges, edgelist.size()/2, &edgelist[0], &edgelen[0], pm);
    double cut_cost;
    if (IS_INT)
      cut_cost = (e_sum + pm_cost)/(-scale2);
    else
      cut_cost = (e_sum + pm_cost)/(-scale);
   return cut_cost;
   //return 0;
  }
    
  
  inline void mexGraph::matching2states(uint32_t* states, double *en, vector<double>& ew, vector<bool>& bicon)
  {
    size_t k;
    int n_nodes = nodes.size();
    int n_edges = ew.size();

    // read matching back, fix edges that are not biconnected
    for (k = 0; k < nedges; ++k)
      edges[k]->agree = edges[k]->other->agree =
	(bicon[k] ? pm->GetSolution(k) : (ew[k] >= 0.0));
    
    std::vector<double> slacks;
    slacks.reserve(nedges);
    for (size_t i = 0; i < nedges; ++i)
      slacks.push_back(pm->ExtendedSlack(i));
    mmc_nodestate(slacks);
    
    // calculate & compare energies
    double lb = 0.0;
    double ub = 0.0;
    for (size_t i = 0; i != n_edges; ++i)
      {
	if (!edges[i]->agree) lb += ew[i];
	if (nodes[edges[i]->node].state !=
	    nodes[edges[i]->other->node].state) ub += ew[i];
      }
    bool same = (std::abs(lb-ub) <= std::abs(lb)*1e-12);
    if (!genus()) assert(same);
    else assert(same || lb <= ub);
    
    // try quenching
    if (!same)
      {
	ub += edge_quench(ew);
	ub += node_quench(ew);
      }

    for (k = 0; k !=(n_nodes-1); ++k)
      states[k] = (nodes[k].state != nodes.back().state ? 1 : 0);
    
    *en = (ub+lb)/2;
  }

  // handle planarity return codes
  inline void mexGraph::errCheck(int k)
  {
    if (k == OK) return;
    
    switch (k)
      {
      case NOTOK:
	edgemap.clear();
	mexErrMsgTxt("something is wrong\n");
	break;
      case NONPLANAR:
	edgemap.clear();
	mexErrMsgTxt("graph is not planar\n");
	break;
      case NONOUTERPLANAR:
	edgemap.clear();
	mexErrMsgTxt("graph is not outerplanar\n");
	break;
      }
  }


}

#endif
